'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var EventEmitter = _interopDefault(require('events'));
var path = require('path');
var fs = require('fs');

function sequence ( array, fn ) {
	var results = [];
	var promise = Promise.resolve();

	function next ( member, i ) {
		return fn( member ).then( function (value) { return results[i] = value; } );
	}

	var loop = function ( i ) {
		promise = promise.then( function () { return next( array[i], i ); } );
	};

	for ( var i = 0; i < array.length; i += 1 ) loop( i );

	return promise.then( function () { return results; } );
}

var opts = { encoding: 'utf-8', persistent: true };

var FileWatcher = function FileWatcher ( file, data, callback, dispose ) {
	try {
		var fsWatcher = fs.watch( file, opts, function (event) {
			if ( event === 'rename' ) {
				fsWatcher.close();
				dispose();
				callback();
			} else {
				// this is necessary because we get duplicate events...
				var contents = fs.readFileSync( file, 'utf-8' );
				if ( contents !== data ) {
					data = contents;
					callback();
				}
			}
		});

		this.fileExists = true;
	} catch ( err ) {
		if ( err.code === 'ENOENT' ) {
			// can't watch files that don't exist (e.g. injected
			// by plugins somehow)
			this.fileExists = false;
		} else {
			throw err;
		}
	}
};

function watch$1 ( rollup, options ) {
	var emitter = new EventEmitter();

	var dests = options.dest ? [ path.resolve( options.dest ) ] : options.target.map( function (target) { return path.resolve( target.dest ); } );
	var filewatchers = new Map();

	var rebuildScheduled = false;
	var building = false;
	var watching = false;

	var timeout;
	var cache;

	function triggerRebuild () {
		clearTimeout( timeout );
		rebuildScheduled = true;

		timeout = setTimeout( function () {
			if ( !building ) {
				rebuildScheduled = false;
				build();
			}
		}, 50 );
	}

	function build () {
		if ( building ) { return; }

		var start = Date.now();
		var initial = !watching;
		if ( cache ) { options.cache = cache; }

		emitter.emit( 'event', { code: 'BUILD_START' });

		building = true;

		return rollup.rollup( options )
			.then( function (bundle) {
				// Save off bundle for re-use later
				cache = bundle;

				bundle.modules.forEach( function (module) {
					var id = module.id;

					// skip plugin helper modules
					if ( /\0/.test( id ) ) { return; }

					if ( ~dests.indexOf( id ) ) {
						throw new Error( 'Cannot import the generated bundle' );
					}

					if ( !filewatchers.has( id ) ) {
						var watcher = new FileWatcher( id, module.originalCode, triggerRebuild, function () {
							filewatchers.delete( id );
						});

						if ( watcher.fileExists ) { filewatchers.set( id, watcher ); }
					}
				});

				// Now we're watching
				watching = true;

				if ( options.targets ) {
					return sequence( options.targets, function (target) {
						var mergedOptions = Object.assign( {}, options, target );
						return bundle.write( mergedOptions );
					});
				}

				return bundle.write( options );
			})
			.then( function () {
				emitter.emit( 'event', {
					code: 'BUILD_END',
					duration: Date.now() - start,
					initial: initial
				});
			}, function (error) {
				emitter.emit( 'event', {
					code: 'ERROR',
					error: error
				});
			})
			.then( function () {
				building = false;
				if ( rebuildScheduled ) { build(); }
			});
	}

	// build on next tick, so consumers can listen for BUILD_START
	process.nextTick( build );

	return emitter;
}

module.exports = watch$1;
